
//MII ethernet interface receiver

module eth_receiver(
	input wire rtl_clk,			//25Mhz clock from eth phy
	input wire hi_clk,
	input wire [3:0]rtl_rxd, 	//received data nibble from eth phy
	input wire rtl_rxdv, 		//received data valid from eth phy
	
	output wire [7:0]rbyte,
	output wire rbyte_valid,
	output wire [10:0]rbyte_addr,
	output  reg [31:0]crc32,
	output wire crc32_ok,
	output wire clk
);

reg [1:0]clkf;
always @(posedge hi_clk)
	clkf <= { clkf[0], rtl_clk };
assign clk = clkf[1];

//fix incoming RTL signals..
reg rxdv_;
reg rxdv;
reg [3:0]rxd;
always @(posedge clk)
begin
	rxdv_ <= rxdv;
	rxdv  <= rtl_rxdv;
	rxd   <= rtl_rxd;
end
		
//catch incoming nibles into 32bit reg
//LSB (Least Significant Bit) goes first
reg [31:0]receiver_reg32;
always @(posedge clk)
	if(rxdv)
		receiver_reg32 <= { rxd, receiver_reg32[31:4] };

assign rbyte = receiver_reg32[31:24];

//reverse bits
wire [7:0]rrbyte; assign rrbyte = {rbyte[0],rbyte[1],rbyte[2],rbyte[3],rbyte[4],rbyte[5],rbyte[6],rbyte[7]};

//count received nibles
reg [11:0]ncounter;
always @(posedge clk)
	if(~rxdv_)
		ncounter <= 0;
	else
	if(rxdv_)
		ncounter <= ncounter + 1;

assign rbyte_addr = ncounter[11:1];
	
assign rbyte_valid = ncounter[0] & rxdv_;

reg [31:0]crc32_;
always @(posedge clk)
	if(rbyte_valid)
	begin
		if(ncounter<16) //  
			crc32_ <= 32'hFFFFFFFF;
		else
			crc32_ <= nextCRC32_D8( rrbyte , crc32_ );
	end

//reverse bits of CRC32
integer i;
always @*
begin
	for ( i=0; i < 32; i=i+1 )
		crc32[i] = crc32_[31-i];
end

reg [31:0]crc32_0;
reg [31:0]crc32_1;
reg [31:0]crc32_2;
reg [31:0]crc32_3;
always @(posedge clk)
	if(rbyte_valid)
	begin
		crc32_3 <= crc32_2;
		crc32_2 <= crc32_1;
		crc32_1 <= crc32_0;
		crc32_0 <= crc32;
	end

reg crc_chk_imp;
always @(posedge clk)
	crc_chk_imp <= rxdv_ & (~rxdv);
	
assign crc32_ok = (&(crc32_3^receiver_reg32)) & crc_chk_imp;

//crc32 function

  // polynomial: (0 1 2 4 5 7 8 10 11 12 16 22 23 26 32)
  // data width: 8
  // convention: the first serial bit is D[7]
  function [31:0] nextCRC32_D8;

    input [7:0] Data;
    input [31:0] crc;
    reg [7:0] d;
    reg [31:0] c;
    reg [31:0] newcrc;
  begin
    d = Data;
    c = crc;

    newcrc[0] = d[6] ^ d[0] ^ c[24] ^ c[30];
    newcrc[1] = d[7] ^ d[6] ^ d[1] ^ d[0] ^ c[24] ^ c[25] ^ c[30] ^ c[31];
    newcrc[2] = d[7] ^ d[6] ^ d[2] ^ d[1] ^ d[0] ^ c[24] ^ c[25] ^ c[26] ^ c[30] ^ c[31];
    newcrc[3] = d[7] ^ d[3] ^ d[2] ^ d[1] ^ c[25] ^ c[26] ^ c[27] ^ c[31];
    newcrc[4] = d[6] ^ d[4] ^ d[3] ^ d[2] ^ d[0] ^ c[24] ^ c[26] ^ c[27] ^ c[28] ^ c[30];
    newcrc[5] = d[7] ^ d[6] ^ d[5] ^ d[4] ^ d[3] ^ d[1] ^ d[0] ^ c[24] ^ c[25] ^ c[27] ^ c[28] ^ c[29] ^ c[30] ^ c[31];
    newcrc[6] = d[7] ^ d[6] ^ d[5] ^ d[4] ^ d[2] ^ d[1] ^ c[25] ^ c[26] ^ c[28] ^ c[29] ^ c[30] ^ c[31];
    newcrc[7] = d[7] ^ d[5] ^ d[3] ^ d[2] ^ d[0] ^ c[24] ^ c[26] ^ c[27] ^ c[29] ^ c[31];
    newcrc[8] = d[4] ^ d[3] ^ d[1] ^ d[0] ^ c[0] ^ c[24] ^ c[25] ^ c[27] ^ c[28];
    newcrc[9] = d[5] ^ d[4] ^ d[2] ^ d[1] ^ c[1] ^ c[25] ^ c[26] ^ c[28] ^ c[29];
    newcrc[10] = d[5] ^ d[3] ^ d[2] ^ d[0] ^ c[2] ^ c[24] ^ c[26] ^ c[27] ^ c[29];
    newcrc[11] = d[4] ^ d[3] ^ d[1] ^ d[0] ^ c[3] ^ c[24] ^ c[25] ^ c[27] ^ c[28];
    newcrc[12] = d[6] ^ d[5] ^ d[4] ^ d[2] ^ d[1] ^ d[0] ^ c[4] ^ c[24] ^ c[25] ^ c[26] ^ c[28] ^ c[29] ^ c[30];
    newcrc[13] = d[7] ^ d[6] ^ d[5] ^ d[3] ^ d[2] ^ d[1] ^ c[5] ^ c[25] ^ c[26] ^ c[27] ^ c[29] ^ c[30] ^ c[31];
    newcrc[14] = d[7] ^ d[6] ^ d[4] ^ d[3] ^ d[2] ^ c[6] ^ c[26] ^ c[27] ^ c[28] ^ c[30] ^ c[31];
    newcrc[15] = d[7] ^ d[5] ^ d[4] ^ d[3] ^ c[7] ^ c[27] ^ c[28] ^ c[29] ^ c[31];
    newcrc[16] = d[5] ^ d[4] ^ d[0] ^ c[8] ^ c[24] ^ c[28] ^ c[29];
    newcrc[17] = d[6] ^ d[5] ^ d[1] ^ c[9] ^ c[25] ^ c[29] ^ c[30];
    newcrc[18] = d[7] ^ d[6] ^ d[2] ^ c[10] ^ c[26] ^ c[30] ^ c[31];
    newcrc[19] = d[7] ^ d[3] ^ c[11] ^ c[27] ^ c[31];
    newcrc[20] = d[4] ^ c[12] ^ c[28];
    newcrc[21] = d[5] ^ c[13] ^ c[29];
    newcrc[22] = d[0] ^ c[14] ^ c[24];
    newcrc[23] = d[6] ^ d[1] ^ d[0] ^ c[15] ^ c[24] ^ c[25] ^ c[30];
    newcrc[24] = d[7] ^ d[2] ^ d[1] ^ c[16] ^ c[25] ^ c[26] ^ c[31];
    newcrc[25] = d[3] ^ d[2] ^ c[17] ^ c[26] ^ c[27];
    newcrc[26] = d[6] ^ d[4] ^ d[3] ^ d[0] ^ c[18] ^ c[24] ^ c[27] ^ c[28] ^ c[30];
    newcrc[27] = d[7] ^ d[5] ^ d[4] ^ d[1] ^ c[19] ^ c[25] ^ c[28] ^ c[29] ^ c[31];
    newcrc[28] = d[6] ^ d[5] ^ d[2] ^ c[20] ^ c[26] ^ c[29] ^ c[30];
    newcrc[29] = d[7] ^ d[6] ^ d[3] ^ c[21] ^ c[27] ^ c[30] ^ c[31];
    newcrc[30] = d[7] ^ d[4] ^ c[22] ^ c[28] ^ c[31];
    newcrc[31] = d[5] ^ c[23] ^ c[29];
    nextCRC32_D8 = newcrc;
  end
  endfunction

endmodule
